/*============================================================================
 * eps.cu
 *      Code to handle eps fitting.
 *==========================================================================*/

#include "math.h"
#include "eps.h"
#include "log.h"
#include "basic_types.h"
#include "basic_ops.h"
#include "mem.h"

#ifdef __NVCC__
#include  <thrust/complex.h>
#endif

#include <string.h>

// Threads in a block for CUDA processing
#define CUDA_BLOCKSIZE          128

#define CUDA_MEM_LIMIT_TEST     (256*1024*1024)

void eps_write_eps(sim_def *sim, const char *fname, Cplx *eps_r)
{
    Uint ii;
    FILE *f;
    
    if ((f = fopen(fname, "wb")) == 0)
    {
        log_printf(log_error, "Unable to open eps output file '%s'.\n", fname);
        return;
    }

    for (ii=0; ii<sim->band.N; ii++)
    {
        fprintf(f, "%le %le\n",
                REAL(eps_r[ii]), IMAG(eps_r[ii]));
    }

    fclose(f);
}

void write_reals(const char *fname, Real *vals, Uint N)
{
    Uint ii;
    FILE *f;
    
    if ((f = fopen(fname, "wb")) == 0)
    {
        log_printf(log_error, "Unable to open eps output file '%s'.\n", fname);
        return;
    }

    for (ii=0; ii<N; ii++)
    {
        fprintf(f, "%e\n", vals[ii]);
    }

    fclose(f);
}

inline __device__ __host__ void get_idx_parm1(Uint &idx_var, parm_sweep_def *parm, Real &dest)
{
    dest = parm->vals[idx_var & parm->N];
    idx_var /= parm->N;
}

#define GET_IDX_PARM1(idx_var, parm, dest)     \
    dest = parm.vals[idx_var % parm.N];      \
    idx_var /= parm.N

/* GPU-based code. */
#ifdef __NVCC__

// eps_fit_gpu_k1
//      First pass of solver, where constant EPS_R is assumed for the full band.
//
__global__ void eps_fit_gpu_k1(sim_def *sim, Cplx *S, Uint f1i, Uint f2i, Real *err_out, Uint N_total)
{
    Real err, f;
    Cplx eps1, sqrt_eps1;
    Cplx S_fit;
    Uint ii, n;
    
    Uint case_idx = blockIdx.x * blockDim.x + threadIdx.x;
    Uint stride = blockDim.x * gridDim.x;

    Uint idx1;
    Real eps_real, eps_imag;

    for (n=case_idx; n<N_total; n+=stride)
    {
        idx1 = n;
        GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
        GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);

        // Compute eps_c assuming constant across the band
        eps1 = Cplx(eps_real, -eps_imag);
        sqrt_eps1 = SQRTC(eps1);
        
        // Compute error for this case at each frequency in range
        err = 0.0;
        for (ii=f1i; ii<=f2i; ii++)
        {
            // Compute the response at this frequency

            // Frequency to check
            f = (Real)ii*sim->df + sim->f0;

            switch (sim->spread_type)
            {
                case 0:
                    get_S_tline1_1d(sim, sqrt_eps1, f, &S_fit);
                    break;
                case 1:
                    get_S_tline1_3d(sim, sqrt_eps1, f, &S_fit);
                    break;
                case 2:
                    get_S_tline1_2d(sim, sqrt_eps1, f, &S_fit);
                    break;
            }
            err += SQ(ABSC(S[ii]-S_fit));
        }
        err_out[n] = err;
    }
}

// eps_fit_gpu_k2
//      Later passes, where a single frequency point is updated at a time to improve fit.
//
__global__ void eps_fit_gpu_k2(sim_def *sim, Cplx *S, Cplx *eps_r_in, Uint band_idx1, Uint band_idx2, Uint band_idx_mod, Real *err_out, Uint N_total)
{
    Real err, f;
    Cplx eps1, eps2, sqrt_eps2;
    Cplx S_fit;
    Uint ii, n;
    Uint fi;
    
    Uint case_idx = blockIdx.x * blockDim.x + threadIdx.x;
    Uint stride = blockDim.x * gridDim.x;

    Uint idx1;
    Real eps_real, eps_imag;
    Cplx eps_l, eps_r;
    
    for (n=case_idx; n<N_total; n+=stride)
    {
        idx1 = n;
        GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
        GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);

        // Get EPS to use for modified band bin
        eps1 = Cplx(eps_real, -eps_imag);
        
        // Loop over bands
        err = 0.0;
        for (ii=band_idx1; ii<band_idx2; ii++)
        {
            // Get complex eps at left and right sample points.  Substitute the
            // trial value if we are at that band index.

            if (ii == band_idx_mod)
            {
                eps_l = eps1;
            }
            else
            {
                eps_l = eps_r_in[ii];
            }

            if ((ii+1) == band_idx_mod)
            {
                eps_r = eps1;
            }
            else
            {
                eps_r = eps_r_in[ii+1];
            }

            // Loop over frequency points in this band
            for (fi = sim->band_fi1[ii]; fi <= sim->band_fi2[ii]; fi++)
            {
                eps2 = eps_l*sim->band_weight_l[fi] + eps_r*sim->band_weight_r[fi];
                sqrt_eps2 = SQRTC(eps2);

                // Frequency to check
                f = (Real)fi*sim->df + sim->f0;

                switch (sim->spread_type)
                {
                    case 0:
                        get_S_tline1_1d(sim, sqrt_eps2, f, &S_fit);
                        break;
                    case 1:
                        get_S_tline1_3d(sim, sqrt_eps2, f, &S_fit);
                        break;
                    case 2:
                        get_S_tline1_2d(sim, sqrt_eps2, f, &S_fit);
                        break;
                }

                err += SQ(ABSC(S[fi]-S_fit));
            }
        }
        err_out[n] = err;
    }
}
#endif

/* CPU-based code. */

void get_S_tline(sim_def *sim, Cplx *eps_all, Cplx *S, Uint f1i, Uint f2i)
{
    const Real c0 = 2.998e8;
    const Cplx eta0 = 376.7;
    
    Uint ii, n;

    Real f;
    Real gn;
    
    Cplx beta, theta;
    Cplx eta;
    Cplx G0;
    //Cplx A;

    Cplx sqrt_eps1;

    Cplx S21_sum, S21_prev;
    Cplx T0;
    
    // Loop through frequencies of interest
    for (ii=f1i; ii<=f2i; ii++)
    {
        f = (Real)ii*sim->df + sim->f0;
        sqrt_eps1 = SQRTC(eps_all[ii-f1i]);

        beta = 2.0*M_PI*f*sqrt_eps1/c0;
        theta = beta*sim->d;
        eta = eta0/sqrt_eps1;

        G0 = (eta-eta0)/(eta+eta0);
        T0 = EXPC(-J*theta);

        // Do n=0 term
        S21_prev = (1.0 + G0)*T0;
        n = 0;
        switch (sim->spread_type)
        {
            case 0:
                gn = 1.0;
                break;
            case 1:
                gn = sim->D/(sim->D + (Real)(2*n+1)*sim->d);
                break;
            case 2:
                gn = SQRT(sim->D/(sim->D + (Real)(2*n+1)*sim->d));
                break;
        }
        S21_sum = S21_prev*(Cplx)gn;

        // Do recursive sum
        for (n=1; n<=sim->ord; n++)
        {
            switch (sim->spread_type)
            {
                case 0:
                    gn = 1.0;
                    break;
                case 1:
                    gn = sim->D/(sim->D + (Real)(2*n+1)*sim->d);
                    break;
                case 2:
                    gn = SQRT(sim->D/(sim->D + (Real)(2*n+1)*sim->d));
                    break;
            }
            S21_prev = S21_prev*G0*G0*T0*T0;
            S21_sum += S21_prev*gn;
        }

        S[ii] = S21_sum*(1.0 - G0);
    }

}

// First pass optimizer.  Assume constant eps_r for whole band.
void eps_fit_cpu1(sim_def *sim, Cplx *S, Cplx *S_fit, Uint f1i, Uint f2i, Uint *best_idx_out, Real *best_err_out)
{
    Real err, best_err;
    Uint best_idx;
    Cplx eps1, best_eps;
    Uint N_total;
    Uint ii, n;

    Uint idx1;
    Real eps_real, eps_imag;
    Cplx eps_vs_freq[f2i-f1i+1];
    
    N_total = sim->eps_real.N * sim->eps_imag.N;

    // Try all possible values
    best_err = REAL_LARGE;
    //best_idx = 0;
    best_eps = 0.0;
    for (n=0; n<N_total; n++)
    {
        idx1 = n;
        GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
        GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);

        for (ii=f1i; ii<=f2i; ii++)
        {
            // Constant with this model
            eps_vs_freq[ii-f1i] = Cplx(eps_real, -eps_imag);
        }

        // Compute the response at this frequency.
        get_S_tline(sim, eps_vs_freq, S_fit, f1i, f2i);

        // Compute error in this solution
        err = 0.0;
        for (ii=f1i; ii<=f2i; ii++)
        {
            err += SQ(ABSC(S[ii]-S_fit[ii]));
        }
        if (err < best_err)
        {
            best_err = err;
            best_idx = n;
        }
    }

    *best_idx_out = best_idx;
    *best_err_out = best_err;
}


// Second pass optimizer, which operates on a band range, and moving just one band point.
void eps_fit_cpu2(sim_def *sim, Cplx *S, Cplx *S_fit, Cplx *eps_r_in, Uint band_idx1, Uint band_idx2, Uint band_idx_mod, Uint *best_idx_out, Real *best_err_out, Real *all_err_out)
{
    Real err, best_err;
    Uint best_idx;
    Cplx eps1, best_eps;
    Uint N_total;
    Uint ii, n;

    Uint idx1, f1, f2, fi;
    Real eps_real, eps_imag;
    Cplx eps_l, eps_r;
    
    Cplx eps_vs_freq[sim->Nf];
    
    N_total = sim->eps_real.N * sim->eps_imag.N;

    // Try all possible values
    best_err = REAL_LARGE;
    //best_idx = 0;
    best_eps = 0.0;
    for (n=0; n<N_total; n++)
    {
        idx1 = n;
        GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
        GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);
        eps1 = Cplx(eps_real, -eps_imag);
        
        err = 0.0;
        for (ii=band_idx1; ii<band_idx2; ii++)
        {
            // Get complex eps at left and right sample points.  Substitute the
            // trial value if we are at that band index.
            if (ii == band_idx_mod)
            {
                eps_l = eps1;
            }
            else
            {
                eps_l = eps_r_in[ii];
            }

            if ((ii+1) == band_idx_mod)
            {
                eps_r = eps1;
            }
            else
            {
                eps_r = eps_r_in[ii+1];
            }
            
            // Loop over frequency points in this band
            //err = 0.0;
            f1 = sim->band_fi1[ii];
            f2 = sim->band_fi2[ii];
            for (fi = f1; fi <= f2; fi++)
            {
                eps_vs_freq[fi-f1] = eps_l*sim->band_weight_l[fi] + eps_r*sim->band_weight_r[fi];
            }

            // Compute the response at this frequency.
            get_S_tline(sim, eps_vs_freq, S_fit, f1, f2);

            // Compute error in this solution
            for (fi=f1; fi<=f2; fi++)
            {
                err += SQ(ABSC(S[fi]-S_fit[fi]));
            }
        }

        if (all_err_out)
        {
            all_err_out[n] = err;
        }

        if (err < best_err)
        {
            best_err = err;
            best_idx = n;
        }
    }

    *best_idx_out = best_idx;
    *best_err_out = best_err;
}

void eps_fit(sim_def *sim, Cplx *S)
{
    Uint f1i, f2i, ii, iter;
    
    Cplx *S_fit=0;
    Cplx *eps_r=0;
    Cplx eps_r1;

    // GPU memory
    sim_def *sim_gpu=0;
    Cplx *S_gpu=0;
    Cplx *eps_r_gpu=0;
    Real *err_gpu=0;
    Real *err_copy=0;

    Uint N_process, numBlocks, n;
    Real best_err;
    Uint best_idx;
    Uint idx1;
        
    Real eps_real, eps_imag;

    Real avg_err;
    Real avg_S;
    Real err_out;

    Uint band_i1, band_i2, band_idx_mod;
    Uint N_avg;
    
    /* Compute average squared value of S parameters to compute % error in fit. */
    avg_S = 0.0;
    for (ii=0; ii<sim->Nf; ii++)
    {
        avg_S += SQ(ABSC(S[ii]));
    }
    avg_S /= (Real)sim->Nf;
    
    /* Allocate space for fitted S-parameters. */
    if (mallocHost((void **) &S_fit, sim->Nf*sizeof(Cplx), sim->use_gpu) == BFALSE)
    {
        log_printf(log_error, "Unable to allocate space for fitted S-parameters.\n");
        goto backout_base;
    }
    memset(S_fit, 0, sim->Nf*sizeof(Cplx));

    /* Allocate space for fitted eps_r values. */
    if (mallocHost((void **) &eps_r, sim->band.N*sizeof(Cplx), sim->use_gpu) == BFALSE)
    {
        log_printf(log_error, "Unable to allocate space for fitted S-parameters.\n");
        goto backout_base;
    }
    memset(eps_r, 0, sim->band.N*sizeof(Cplx));
    
    // Precompute temporary data to speed things up
    sim->df = (sim->f1 - sim->f0)/(Real)(sim->Nf - 1);

    N_process = sim->eps_real.N*sim->eps_imag.N;
    
    // If running on the GPU, allocate space and transfer data to GPU
    if (sim->use_gpu)
    {
        numBlocks = (N_process + CUDA_BLOCKSIZE - 1) / CUDA_BLOCKSIZE;
        //printf("numBlocks=%d N_process=%d\n", numBlocks, N_process);
        
#ifdef __NVCC__        
        cudaMalloc(&sim_gpu, sizeof(sim_def));
        if (sim_gpu == 0)
        {
            log_printf(log_error, "Failed to allocate sim_def.\n");
            goto backout_base;
        }
        cudaMemcpy(sim_gpu, sim, sizeof(sim_def), cudaMemcpyHostToDevice);
           
        cudaMalloc(&S_gpu, sim->Nf*sizeof(Cplx));
        if (S_gpu == 0)
        {
            log_printf(log_error, "Failed to allocate S_gpu.\n");
            goto backout_base;
        }
        cudaMemcpy(S_gpu, S, sim->Nf*sizeof(Cplx), cudaMemcpyHostToDevice);

        cudaMalloc(&eps_r_gpu, sim->band.N*sizeof(Cplx));
        if (eps_r_gpu == 0)
        {
            log_printf(log_error, "Failed to allocate eps_r_gpu.\n");
            goto backout_base;
        }
        
        cudaMalloc(&err_gpu, N_process*sizeof(Real));
        if (err_gpu == 0)
        {
            log_printf(log_error, "Failed to allocate err_gpu.\n");
            goto backout_base;
        }

#endif
    }

    if (mallocHost((void **) &err_copy, N_process*sizeof(Real), sim->use_gpu) == BFALSE)
    {
        log_printf(log_error, "Unable to allocate space for err_copy.\n");
        goto backout_base;
    }
    
    // 1. Get optimal solution for constant eps_r across all bands.
    f1i = sim->band_fi1[0];
    f2i = sim->band_fi2[sim->band.N-2];
    if (sim->use_gpu)
    {
        log_printf(log_info, "GPU 1: Computing %d cases.\n", N_process);
        
        eps_fit_gpu_k1<<<numBlocks, CUDA_BLOCKSIZE>>>(sim_gpu, S_gpu, f1i, f2i, err_gpu, N_process);
        cudaDeviceSynchronize();

        // Copy back error results
        cudaMemcpy(err_copy, err_gpu, N_process*sizeof(Real), cudaMemcpyDeviceToHost);
            
        // Find the solution with the best fit over the blocks
        log_printf(log_info, "CPU: Minimizing. ");
            
        best_err = REAL_LARGE;
        best_idx = 0;
        for (n=0; n<N_process; n++)
        {
            if (err_copy[n] < best_err)
            {
                best_err = err_copy[n];
                best_idx = n;
            }
        }
        err_out = best_err;
        log_printf(log_info, "\n");
    }
    else
    {
        log_printf(log_info, "CPU: Computing %d cases. ", N_process);
            
        // Find best fit for this band using GPU
        eps_fit_cpu1(sim, S, S_fit, f1i, f2i, &best_idx, &err_out);
        log_printf(log_info, "\n");
    }

    avg_err = err_out/(Real)(f2i-f1i+1);
    log_printf(log_info, "RMS Error: %.1f %%\n", sqrtf(avg_err/avg_S)*100.0);

    // Get parameter set providing best fit
    idx1 = best_idx;
    GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
    GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);
    log_printf(log_info, "eps_r=%e eps_i=%e\n", eps_real, eps_imag);
    
    // Fill out eps_c for the bands according to the best value obtained
    for (ii=0; ii<sim->band.N; ii++)
    {
        eps_r[ii] = Cplx(eps_real, -eps_imag);
    }

    // Do a certain number of optimization iterations.
    for (iter=0; iter < sim->N_iter; iter++)
    {
        // Loop over band points to change
        avg_err = 0;
        N_avg = 0;
        for (ii=0; ii < sim->band.N; ii++)
        {
            // Get range of bands to optimize based on where we
            // are in the band sequence.
            if (ii==0)
            {
                band_i1 = 0;
                band_i2 = 1;
            }
            else if (ii==(sim->band.N-1))
            {
                band_i1 = ii-1;
                band_i2 = ii;
            }
            else
            {
                band_i1 = ii-1;
                band_i2 = ii+1;
            }
            band_idx_mod = ii;

            // Optimize
            
            if (sim->use_gpu)
            {
                log_printf(log_info, "GPU Iter %d Band %d: Computing %d cases. ", iter, ii, N_process);

                cudaMemcpy(eps_r_gpu, eps_r, (sim->band.N)*sizeof(Cplx), cudaMemcpyHostToDevice);
                
                eps_fit_gpu_k2<<<numBlocks, CUDA_BLOCKSIZE>>>(sim_gpu, S_gpu, eps_r_gpu, band_i1, band_i2, band_idx_mod, err_gpu, N_process);
                cudaDeviceSynchronize();
                
                // Copy back error results
                cudaMemcpy(err_copy, err_gpu, N_process*sizeof(Real), cudaMemcpyDeviceToHost);
                
                // Find the solution with the best fit over the blocks
                log_printf(log_info, "CPU: Minimizing. ");
                
                best_err = REAL_LARGE;
                best_idx = 0;
                for (n=0; n<N_process; n++)
                {
                    if (err_copy[n] < best_err)
                    {
                        best_err = err_copy[n];
                        best_idx = n;
                    }
                }
                //write_reals("err_gpu.out", err_copy, N_process);
                log_printf(log_info, "\n");
            }
            else
            {
                log_printf(log_info, "CPU Iter %d Band %d: Computing %d cases. ", iter, ii, N_process);
                
                // Find best fit for this band using CPU
                eps_fit_cpu2(sim, S, S_fit, eps_r, band_i1, band_i2, band_idx_mod, &best_idx, &best_err, err_copy);
                // DEBUG
                //write_reals("err_cpu.out", err_copy, N_process);
                log_printf(log_info, "\n");
            }
            
            avg_err = avg_err + best_err;
            N_avg += (sim->band_fi2[band_i2-1] - sim->band_fi1[band_i1] + 1);
            
            // Get parameter set providing best fit
            idx1 = best_idx;
            GET_IDX_PARM1(idx1, sim->eps_real, eps_real);
            GET_IDX_PARM1(idx1, sim->eps_imag, eps_imag);
            
            // Update that band index point
            eps_r[band_idx_mod] = Cplx(eps_real, -eps_imag);

        } // ii (bands)
        avg_err /= (Real)N_avg;

        log_printf(log_info, "RMS Error: %.1f %%\n", sqrtf(avg_err/avg_S)*100.0);
        
    } // iter

    // Write fitted eps_r to a file
    eps_write_eps(sim, "eps_r.out", eps_r);

 backout_base:
    if (S_fit) freeHost(S_fit, sim->use_gpu);
    if (eps_r) freeHost(eps_r, sim->use_gpu);
    if (err_copy) freeHost(err_copy, sim->use_gpu);
#ifdef __NVCC__
    if (sim_gpu) cudaFree(sim_gpu);
    if (eps_r_gpu) cudaFree(eps_r_gpu);
    if (S_gpu) cudaFree(S_gpu);
    if (err_gpu) cudaFree(err_gpu);

#endif                          
    
    return;
}

