/*============================================================================
 * sim.cc
 *      Code to handle reading and storing simulation parameters.
 *==========================================================================*/

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "global_defs.h"
#include "sim.h"
#include "basic_ops.h"
#include "log.h"
#include "mem.h"
#include "eps.h"

sim_def sim_parms;

Bool sim_fill_range1(parm_sweep_def *p)
{
    Uint i;

    if (p->N > MAX_PARM_RANGE)
    {
        log_printf(log_error, "Encountered range is too large (%d).  Need to recompile for this.", p->N);
        return(BFALSE);
    }
    
    if (p->N == 1)
    {
        p->vals[i] = p->start;
    }
    else
    {
        for (i=0; i<p->N; i++)
        {
            switch(p->is_log)
            {
                case 0:
                    p->vals[i] = (Real)i*(p->stop - p->start)/((Real)(p->N - 1)) + p->start;
                    break;
                    
                default:
                    p->vals[i] = exp((Real)i*(log(p->stop) - log(p->start))/((Real)(p->N - 1)) + log(p->start));
                    break;
                    
            }

            
        }
    }

    return(BTRUE);
}

Bool sim_fill_range(sim_def *sim)
{
    if (!sim_fill_range1(&(sim->eps_real))) return(BFALSE);
    if (!sim_fill_range1(&(sim->eps_imag))) return(BFALSE);
    if (!sim_fill_range1(&(sim->band))) return(BFALSE);
    return(BTRUE);
}

Bool sim_fill_bands(sim_def *sim)
{
    Uint ii;
    Real f;
    Real f_band_l;
    Real f_band_r;
    Uint fi;
    
    // Set the range of frequency samples for each band
    for (ii=0; ii<sim->band.N-1; ii++)
    {
        f_band_l = sim->band.vals[ii];
        f_band_r = sim->band.vals[ii+1];

        sim->band_fi1[ii] = (Uint)ceilf((f_band_l - sim->f0)/(sim->f1 - sim->f0)*(Real)(sim->Nf-1));
        sim->band_fi2[ii] = (Uint)floorf((f_band_r - sim->f0)/(sim->f1 - sim->f0)*(Real)(sim->Nf-1));
    }

    // Set the interpolation weights and indices
    // Deactivate all frequency samples by default.
    for (ii=0; ii<sim->Nf; ii++)
    {
        sim->band_weight_l[ii] = -1;
        sim->band_weight_r[ii] = -1;
    }

    // Go through each band and assign left and right weights to frequency
    // points.
    for (ii=0; ii<sim->band.N-1; ii++)
    {
        // Loop over frequencies in this band
        for (fi=sim->band_fi1[ii]; fi<=sim->band_fi2[ii]; fi++)
        {
            f = (Real)fi*(sim->f1 - sim->f0)/(Real)sim->Nf + sim->f0;

            sim->band_idx_l[fi] = ii;
            sim->band_idx_r[fi] = ii+1;
            
            // val[f] = (f - f_l)/(f_r - f_l)*(val_r - val_l) + val_l;
            sim->band_weight_r[fi] = (f - sim->band.vals[ii])/(sim->band.vals[ii+1] - sim->band.vals[ii]);
            sim->band_weight_l[fi] = 1.0 - sim->band_weight_r[fi];
        }
    }

    return(BTRUE);
}

Bool sim_read(const char *fname, sim_def *sim)
{
    FILE *f;
    Bool ccode;
    double d1, d2, d3;
    Int i1;
    Uint ii, kk;
    char line1[256];
    parm_sweep_def *p;
    
    // Defaults
    memset(sim, 0, sizeof(sim_def));
    sim->D = REAL_LARGE;
    sim->ord = 10;
    sim->log_level = log_error;
    sim->spread_type = 1;  // Default to 3D
    
    if ((f = fopen(fname, "rb")) == 0)
    {
        log_printf(log_error, "Unable to open parameter file '%s'\n", fname);
        ccode = BFALSE;
        goto backout_base;
    }

    while (fgets(line1, sizeof(line1), f) != 0)
    {
        if ((line1[0] == '#') || (line1[0] == '%') || (line1[0] == '\r') ||
            (line1[0] == '\n') || (line1[0] == 0)) continue;
        
        if (strncmp(line1, "gpu ", 4) == 0)
        {
            if (sscanf(line1+4, "%d", &i1) != 1)
            {
                log_printf(log_error, "Error parsing gpu parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->use_gpu = (Bool) i1;
        }
        else if (strncmp(line1, "D ", 2) == 0)
        {
            if (sscanf(line1+2, "%lf", &d1) != 1)
            {
                log_printf(log_error, "Error parsing D parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->D = d1;
        }
        else if (strncmp(line1, "d ", 2) == 0)
        {
            if (sscanf(line1+2, "%lf", &d1) != 1)
            {
                log_printf(log_error, "Error parsing d parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->d = d1;
        }
        else if (strncmp(line1, "eps_real ", 9) == 0)
        {
            p = &(sim->eps_real);
            if (sscanf(line1+9, "%f %f %d %d", &(p->start), &(p->stop), &(p->N), &(p->is_log)) != 4)
            {
                log_printf(log_error, "Error parsing eps_inf parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
        }
        else if (strncmp(line1, "eps_imag ", 9) == 0)
        {
            p = &(sim->eps_imag);
            if (sscanf(line1+9, "%f %f %d %d", &(p->start), &(p->stop), &(p->N), &(p->is_log)) != 4)
            {
                log_printf(log_error, "Error parsing eps_s parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
        }
        else if (strncmp(line1, "band ", 5) == 0)
        {
            p = &(sim->band);
            if (sscanf(line1+5, "%f %f %d %d", &(p->start), &(p->stop), &(p->N), &(p->is_log)) != 4)
            {
                log_printf(log_error, "Error parsing eps_s parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
        }
        else if (strncmp(line1, "f ", 2) == 0)
        {
            if (sscanf(line1+2, "%lf %lf %d", &d1, &d2, &i1) != 3)
            {
                log_printf(log_error, "Error parsing f parameters.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->f0 = d1;
            sim->f1 = d2;
            sim->Nf = i1;
        }
        else if (strncmp(line1, "ord ", 4) == 0)
        {
            if (sscanf(line1+4, "%d", &i1) != 1)
            {
                log_printf(log_error, "Error parsing ord parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->ord = i1;
        }
        else if (strncmp(line1, "log ", 4) == 0)
        {
            if (sscanf(line1+4, "%d", &i1) != 1)
            {
                log_printf(log_error, "Error parsing log level parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->log_level = (log_level_enum) i1;
        }
        else if (strncmp(line1, "iter ", 5) == 0)
        {
            if (sscanf(line1+5, "%d", &i1) != 1)
            {
                log_printf(log_error, "Error parsing iter parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->N_iter = i1;
        }
        else if (strncmp(line1, "spread ", 7) == 0)
        {
            if (sscanf(line1+7, "%d", &i1) != 1)
            {
                log_printf(log_error, "Error parsing log level parameter.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            if ((i1 < 0) || (i1 > 2))
            {
                log_printf(log_error, "Invalid spread type.\n");
                ccode = BFALSE;
                goto backout_file;
            }
            sim->spread_type = i1;
        }
        else
        {
            log_printf(log_warning, "Warning: Unrecognized parameter line in '%s':\n", fname);
            log_printf(log_warning, "%s\n", line1);
        }
    }

    // Do some rudimentary checking

    if (sim->eps_real.N == 0)
    {
        log_printf(log_error, "eps_real range not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }

    if (sim->eps_imag.N == 0)
    {
        log_printf(log_error, "eps_imag range not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }

    if (sim->Nf == 0)
    {
        log_printf(log_error, "Frequency range not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }

    if (sim->band.N == 0)
    {
        log_printf(log_error, "Fitting bands not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }
    
    if (sim->d == 0.0)
    {
        log_printf(log_error, "Sample thickness not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }
    
    if (sim->N_iter == 0)
    {
        log_printf(log_error, "Number of iterations not set.\n");
        ccode = BFALSE;
        goto backout_file;
    }
    
    fclose(f);
    return(BTRUE);

 backout_file:
    fclose(f);

 backout_base:
    return(ccode);
}

// Reads S21
Cplx *sim_read_S(const char *fname, sim_def *sim)
{
    Cplx *p;
    Uint i;
    double rp, ip;
    FILE *f;
    
    if ((f=fopen(fname, "rb")) == 0)
    {
        log_printf(log_error, "Unable to open S-parameters file '%s'.\n", fname);
        goto backout_base;
    }
    
    if (mallocHost((void **) &p, sim->Nf*sizeof(Cplx), sim->use_gpu) == BFALSE)
    {
        log_printf(log_error, "Unable to allocate space for S-parameters.\n");
        goto backout_file;
    }

    for (i=0; i<sim->Nf; i++)
    {
        if (fscanf(f, "%le %le", &rp, &ip) != 2)
        {
            log_printf(log_error, "Unable to read %d S-parameters from '%s'.\n", sim->Nf, fname);
            goto backout_base;
        }
        p[i] = Cplx(rp, ip);
    }

    fclose(f);
    return(p);
    
 backout_mem:
    freeHost(p, sim->use_gpu);

 backout_file:
    fclose(f);
    
 backout_base:
    return(0);
}

Bool sim_run(const char *parm_file, const char *s_parm_file)
{
    Bool ccode;
    Uint ii, kk;
    Cplx *S;
    
    ccode = BFALSE;
    
    // Read top-level parameters
    if (sim_read(parm_file, &sim_parms) == BFALSE)
    {
        goto backout_base;
    }
    log_init(sim_parms.log_level);

    // Fill in ranges
    if (sim_fill_range(&sim_parms) == BFALSE)
    {
        goto backout_base;
    }

    log_printf(log_info, "Material Dielectric Fitting Code (Iterative1).  Version: %d.%d\n", VERSION_MAJOR, VERSION_MINOR);
    
    if (sim_parms.use_gpu)
    {
        log_printf(log_info, "CUDA GPU acceleration enabled\n");
    }
    else
    {
        log_printf(log_info, "Running on CPU only (no GPU acceleration)\n");
    }

    switch (sim_parms.spread_type)
    {
        case 0:
            log_printf(log_info, "No sppreading (1D).\n");
            break;
        case 1:
            log_printf(log_info, "3D spreading.\n");
            break;
        case 2:
            log_printf(log_info, "2D spreading.\n");
            break;
    }
    
    if ((S = sim_read_S(s_parm_file, &sim_parms)) == 0)
    {
        goto backout_base;
    }

    // Compute band weights for frequency points.
    if (sim_fill_bands(&sim_parms) == BFALSE)
    {
        goto backout_base;
    }
    
    eps_fit(&sim_parms, S);
    
    log_printf(log_info, "Done.\n");
    
    // Success
    ccode = BTRUE;

 backout_S:
    freeHost(S, sim_parms.use_gpu);
    
 backout_base:
    return(ccode);

}

